/*------------------------------------------------------------------------
 
   File:              analyseerlogmanfile.p
 
   Description:       Identifies possible memory leaks and their source by
                      extracting those lines from logmanager output
                      that point at created objects that are not deleted later on
                      or allocated space that is not (fully) deallocated.
                      
                      Create the logmanagerfile by adding these parameters to
                      your pf (or have a look at the help for the log-manager ABL keyword and use
                      the log-manager atributes and methods):
                      
                      -clientlog <path + filename where you want the logmanager output written to>
                      -logginglevel 4
                      -clearlog
                      -logentrytypes dynobjects.*
                       
                      Further info about dynamic object logging can be found on
                      http://www.progress.com/progress/products/documentation/docs/dvdbg/dvdbg.pdf 

   Input Parameters:  picInputFrom: logmanager output filename (see -clientlog)
                      picOutputTo : output filename 
 
   Output Parameters:
       <none>
 
   Author:            Stefan Houtzager 
 
   Created:           January 6 2010 
                      Houtzager ICT consultancy & development
                      
   Notes:             Written with 10.2A02. This analyser accounts for garbagecollection.
                      
                      Note that some memoryleaks are caused bij others. If for example
                      you have an instance of a class defined with USE-WIDGET-POOL and
                      that instance is nor explicitly deleted nor garbage-collected
                      while it should have been you will find besides a line for the created
                      object a line in the logmanager output about creating a widget-pool: 
                      'Creating pool', but none about deleting that pool: 'Deleting pool'.
                      
                      comments/bugreports welcome, send them to agent_008_nl@yahoo.com   
   
   Revisions:                   
 ------------------------------------------------------------------------*/  
  ROUTINE-LEVEL ON ERROR UNDO, THROW. 
  /*  output van log-manager analyseren; uitfilteren van niet verwijderde dynamische objecten
      die mogelijk een geheugenlek veroorzaken */
  DEFINE INPUT PARAMETER picInputFrom AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER picOutputTo  AS CHARACTER NO-UNDO.

  DEFINE VARIABLE cLine                   AS CHARACTER  NO-UNDO.
  DEFINE VARIABLE cLineWithoutExtraGaps   AS CHARACTER  NO-UNDO.
  DEFINE VARIABLE iLineNr                 AS INTEGER    NO-UNDO.
  DEFINE VARIABLE iActivity                 AS INTEGER    NO-UNDO. 

  /* All DynObjects except LONGCHAR  MEMPTR and widget-pools */
  DEFINE TEMP-TABLE ttLines NO-UNDO
      FIELD Line    AS CHARACTER
      FIELD Created AS LOGICAL
      FIELD Hdl     AS INTEGER
      FIELD ObjType AS CHARACTER
      INDEX PU_Lines IS PRIMARY UNIQUE Created  Hdl.
  
  /* LONGCHAR  MEMPTR */
  DEFINE TEMP-TABLE ttAllocLines NO-UNDO
      FIELD Line      AS CHARACTER
      FIELD Allocated AS LOGICAL
      FIELD SIZE      AS INTEGER
      FIELD Linenr    AS INTEGER
      FIELD ObjType   AS CHARACTER
      FIELD Adress    AS CHARACTER 
      INDEX PU_LinesAlloc IS PRIMARY UNIQUE Allocated Adress Linenr. /* The same adress can be reused after deallocation */
  
  /* widget-pools */
  DEFINE TEMP-TABLE ttPoolLines NO-UNDO
      FIELD Line      AS CHARACTER
      FIELD Created   AS LOGICAL
      FIELD Substr80  AS CHARACTER 
      INDEX PU_LinesAlloc IS PRIMARY Created Substr80.

  DEFINE BUFFER bttLines      FOR ttLines.
  DEFINE BUFFER bttAllocLines FOR ttAllocLines.
  DEFINE BUFFER bttPoolLines  FOR ttPoolLines.  

  &SCOPED-DEFINE CREATED 1    
  &SCOPED-DEFINE DELETED 2    
  &SCOPED-DEFINE ALLOCATED 3 
  &SCOPED-DEFINE DEALLOCATED 4 
  &SCOPED-DEFINE CREATEPOOL 5
  &SCOPED-DEFINE DELETEPOOL 6

  INPUT FROM VALUE(picInputFrom).

  REPEAT ON ERROR UNDO, THROW:

    IMPORT UNFORMATTED cLine.

    IF ENTRY(6,cLine,' ') <> "DYNOBJECTS" THEN NEXT.

    ASSIGN iLineNr = iLineNr + 1
           cLineWithoutExtraGaps = cLine. 

    IF ENTRY(7,cLine,' ')                           BEGINS 'Created' /* created of created-ref */   THEN iActivity = {&CREATED}.
    ELSE IF ENTRY(7,cLine,' ')                      BEGINS 'Deleted' /* Deleted of deleted-by-gc */ THEN iActivity = {&DELETED}.
    ELSE IF ENTRY(7,cLine,' ')                      =      'Allocated' /*  LONGCHAR  MEMPTR */     THEN iActivity = {&ALLOCATED}.
    ELSE IF ENTRY(7,cLine,' ')                      =      'Deallocated' /*  LONGCHAR  MEMPTR */   THEN iActivity = {&DEALLOCATED}.
    ELSE IF ENTRY(7,cLine,' ') + ENTRY(8,cLine,' ') =      'Creatingpool'                           THEN iActivity = {&CREATEPOOL}.
    ELSE IF ENTRY(7,cLine,' ') + ENTRY(8,cLine,' ') =      'Deletingpool'                           THEN iActivity = {&DELETEPOOL}.
    ELSE iActivity = 0.

    IF iActivity = 0 THEN NEXT.

    DO WHILE INDEX(cLineWithoutExtraGaps,'  ') > 0:
       cLineWithoutExtraGaps = REPLACE(cLineWithoutExtraGaps,'  ',' ').
    END.

    IF iActivity = {&CREATED} OR iActivity = {&DELETED} THEN DO:
      CREATE ttLines.        
      ASSIGN ttLines.Line    = cLine
             ttLines.Created = iActivity = {&CREATED}
             ttLines.ObjType = IF SUBSTRING(cLine,80,11) = '.NET Object' THEN '.NET Object' ELSE ENTRY(8,cLineWithoutExtraGaps,' ')
             ttLines.Hdl     = INTEGER(ENTRY(2,ENTRY(IF ttLines.ObjType NE '.NET Object' THEN 9 ELSE 10,cLineWithoutExtraGaps,' '),':')).
    END.
    ELSE IF iActivity = {&ALLOCATED} OR iActivity = {&DEALLOCATED} THEN DO:

      IF ENTRY(9,cLineWithoutExtraGaps,' ') = 'handle:0' THEN NEXT. /* no need for this line */

      CREATE ttAllocLines.        
      ASSIGN ttAllocLines.Line      = cLine
             ttAllocLines.Linenr    = iLineNr
             ttAllocLines.Allocated = iActivity = {&ALLOCATED}
             ttAllocLines.ObjType   = ENTRY(8,cLineWithoutExtraGaps,' ') 
             ttAllocLines.Adress    = ENTRY(9,cLineWithoutExtraGaps,' ')
             ttAllocLines.SIZE      = INTEGER(ENTRY(11,cLineWithoutExtraGaps,' ')).
    END.
    ELSE /* {&CREATEPOOL} or {&DELETEPOOL} */ DO:
      CREATE ttPoolLines.
      ASSIGN ttPoolLines.Line      = cLine   
             ttPoolLines.Created   = iActivity = {&CREATEPOOL}
             ttPoolLines.Substr80  = SUBSTRING(cLine, 80) /* line contents after 'Creating pool' or 'Deleting pool' */ .
    END.

  END. /* REPEAT */

  OUTPUT TO VALUE(picOutputTo).

  /* All DynObjects except LONGCHAR  MEMPTR and pools */
  FOR EACH ttLines
       WHERE ttLines.Created = TRUE
       ON ERROR UNDO, THROW:

    FIND bttLines 
        WHERE bttLines.Created  = FALSE 
          AND bttLines.Hdl      = ttLines.Hdl NO-ERROR.

    IF NOT AVAILABLE bttLines THEN PUT UNFORMATTED ttLines.Line SKIP.
  END.

  /* LONGCHAR  MEMPTR */
  FOR EACH ttAllocLines 
      WHERE ttAllocLines.Allocated = TRUE
      ON ERROR UNDO, THROW:

    FIND FIRST bttAllocLines 
        WHERE bttAllocLines.Allocated = FALSE 
          AND bttAllocLines.Adress    = ttAllocLines.Adress
          AND bttAllocLines.Linenr    > ttAllocLines.Linenr NO-ERROR.

    IF NOT AVAILABLE bttAllocLines THEN PUT UNFORMATTED ttAllocLines.Line SKIP.

    /* if the object is not fully deallocated add a Line with the deallocationsize */
    IF AVAILABLE bttAllocLines 
     AND bttAllocLines.SIZE NE ttAllocLines.SIZE 
        THEN PUT UNFORMATTED ttAllocLines.Line + " (" + STRING(bttAllocLines.SIZE) + " deallocated!)"  SKIP.
  END.

  /* pools */
  FOR EACH ttPoolLines
       WHERE ttPoolLines.Created = TRUE
       ON ERROR UNDO, THROW:

    FIND FIRST bttPoolLines 
        WHERE bttPoolLines.Created   = FALSE 
          AND bttPoolLines.Substr80  = ttPoolLines.Substr80 NO-ERROR.

    IF NOT AVAILABLE bttPoolLines THEN PUT UNFORMATTED ttPoolLines.Line SKIP.
    ELSE DELETE bttPoolLines. /*  should not be found more than once */
  END.

  CATCH ProError AS Progress.Lang.ProError: 
    MESSAGE 'Error, cannot proceed creating ' + picOutputTo + '~n' +  ProError:GetMessage(1) VIEW-AS ALERT-BOX ERROR.
    DELETE OBJECT ProError. /* cannot afford to create leaks myself ;-) */
  END CATCH.

  FINALLY: /* to be sure all streams are closed */
    INPUT CLOSE.
    OUTPUT CLOSE.
  END FINALLY.
